import time
import datetime
import timeit
import os
import io
import serial
from serial.tools.list_ports import comports
from serial.serialutil import SerialException
from xmodem import XMODEM1k


# ASCII Control Codes
# -------------------

# terminates command output send from board
EOT = 0x04

# used to sync communication, board replies with SYN when it receives one
SYN = 0x16


class Reaper(object):
    def __init__(self, port=None, data_dir=None, timeout=0.5, echo=True):
        self.port = port
        self.data_dir = data_dir
        self.timeout = timeout
        self.conn = None
        self.echo = echo
        self._devices_by_id = None
        self._devices_by_name = {}

        if data_dir and not os.path.exists(self.data_dir):
            os.makedirs(self.data_dir)

    def pr(self, s, end='\n', flush=False):
        if self.echo:
            print(s, end=end, flush=flush)

    @staticmethod
    def find():
        """Find devices running Arduino Reaper.

        This function enumerates the USB ports on the system and tries to
        connect to each and get device information.

        Returns list of port names that appear to be Arduino Reaper devices.
        """
        def is_reaper_device(port):
            try:
                reaper = Reaper(port=port, echo=False)
                return reaper.connect() and reaper.info(timeout=5)
            except Exception as e:
                return False

        # only consider ports that appear to be USB devices
        usb_ports = [p.device for p in comports() if p.usb_description()]

        # return reaper devices
        return [p for p in usb_ports if is_reaper_device(p)]

    def connect(self, retry_delay=1.0, max_tries=None):
        """Connect to serial port, waiting until it is ready"""
        self.pr('connecting to port {} ...'.format(self.port),
                end='',
                flush=True)

        tries = 0
        while self.conn is None and (max_tries is None or tries < max_tries):
            tries += 1
            try:
                self.conn = serial.Serial(port=self.port, timeout=self.timeout)
                self.pr(' connected')
                time.sleep(.1)
                return True
            except SerialException as se:
                time.sleep(retry_delay)
                self.pr('.', end='.', flush=True)
        return False

    def disconnect(self):
        if self.conn:
            self.pr('disconnecting {}'.format(self.conn.name))
            self.conn.close()

    def read(self, timeout=None, stringify=False):
        """Read until EOT and return data read, not including EOT

        :param timeout: how long to wait for response to complete
        :param stringify: if True return a string from ASCII bytes, else
            return the bytes as is
        """
        data = []
        t1 = time.time()
        t2 = t1

        while True:
            t2 = time.time()
            if timeout is not None and t2 - t1 > timeout:
                break
            bytesRead = self.conn.read(size=1)
            if not bytesRead or bytesRead[0] == EOT:
                break
            self.pr(chr(bytesRead[0]), end='', flush=True)
            data.append(bytesRead[0])

        if stringify:
            return ''.join([chr(b) for b in data])
        else:
            return data

    def sync(self, timeout=None):
        t1 = time.time()
        t2 = t1
        while True:
            t2 = time.time()
            if timeout is not None and t2 - t1 > timeout:
                raise TimeoutError()
            self.conn.write([SYN])
            time.sleep(.01)
            response = self.read(timeout=1)
            if response and response[0] == SYN:
                return
            elif response:
                self.pr(response, flush=True)

    def commands(self, commands):
        """Run a list of commands and return results"""
        self.read(timeout=3)
        results = []
        for cmd in commands:
            results.append(self.command(cmd))
        return results

    def command(self, command):
        """Run a single command and return results"""
        parts = command.split()
        verb = parts[0]
        if verb == 'cp':
            # get file size
            filename = parts[1]
            self.send_command('ls {}'.format(filename))
            file_info = self.read(stringify=True)
            size = int(file_info.split()[2])

            # receive file
            path = filename.split('/')
            local_filename = os.path.join('.', *path)
            self.send_command(command)

            def progress_fun(num_bytes, w):
                print('.', end='', flush=True)

            prev_echo = self.echo
            self.echo = False
            self.receiveFile(local_filename, size, progress_fun)
            self.echo = prev_echo

            self.pr('received {}, size={}'.format(filename, size))
        elif verb == 'info':
            self.send_command(command)
            return self.read()
        else:
            self.send_command(command)
            return self.read()

    def info(self, timeout=None):
        self.send_command('info', timeout=timeout)
        fields = {}
        response = self.read(stringify=True, timeout=timeout) or ''
        for line in response.splitlines():
            name, value = line.split('\t')
            fields[name] = value
        return fields

    def rm(self, sd_filename):
        self.send_command('rm {}'.format(sd_filename))
        return self.read(stringify=True)

    def cp(self, sd_filename, local_filename, size, progress_fun=None):
        self.send_command('cp {}'.format(sd_filename))
        self.receiveFile(local_filename, size, progress_fun)

    def set_time(self, time=None):
        """Set device time in UTC"""
        if time is None:
            time = datetime.datetime.utcnow()
        self.send_command('set_time {} {} {} {} {} {}'.format(time.year - 2000,
                                                              time.month,
                                                              time.day,
                                                              time.hour,
                                                              time.minute,
                                                              time.second))
        return self.read(stringify=True)

    def get_time(self):
        """Get device time, assumed to be in UTC"""
        self.send_command('get_time')

        # 6 unsigned ints in range (0-255)
        # year - 0 = year 2000
        # month - 1-12
        # day 1-31
        # minute 0-59
        # second 0-60 (could theoretically have a leap second, but this
        #              will not happen unless set_time is used with a
        #              leap second and then get_time completes within
        #              the same second)
        time_string = self.read(stringify=True)
        year, month, day, hour, minute, second = map(int, time_string.split())
        dt = datetime.datetime(year + 2000, month, day, hour, minute, second,
                               tzinfo=datetime.timezone.utc)
        return dt

    def ls(self):
        """Get list of files"""
        self.send_command('ls')
        result = []
        listing = self.read(stringify=True)

        current_dir = []
        prev_level = 0
        for line in listing.splitlines()[1:]:
            level = int((len(line) - len(line.lstrip())) / 2)
            date, time, size, name = line.split()

            if name.endswith('/'):
                current_dir[level:] = [name]
            else:
                current_dir = current_dir[0:level]
                result.append(dict(
                    name=''.join(current_dir + [name]),
                    size=int(size),
                    last_modified=' '.join([date, time])))

            name = current_dir
            prev_level = level
        return result

    def send_command(self, command, timeout=None):
        self.sync(timeout=timeout)
        time.sleep(.01)
        for c in command.rstrip('\n'):
            self.conn.write([ord(c)])
            time.sleep(.01)
        self.conn.write([ord('\n')])
        self.conn.flush()
        time.sleep(.01)

    def receiveFile(self, filename, size, progress_fun):
        def getc(size, timeout=1):
            self.conn.timeout = timeout
            data = self.conn.read(size)
            if len(data) == 0:
                data = None
            return data

        def putc(data, timeout=1):
            return self.conn.write(data)

        xm = XMODEM1k(getc, putc)
        with open(filename, 'wb+') as f:
            stream = StreamWriter(f, progress_fun)
            xm.recv(stream)
            report = stream.report()
            f.truncate(size)
            self.conn.timeout = self.timeout
        return report

    def set_device_name(self, name, device_id):
        if self._devices_by_id is None:
            self._read_devices()
        if self._devices_by_name.get(name):
            return False

        self._devices_by_id[device_id] = name
        self._devices_by_name[name] = device_id
        self._write_devices()
        return True

    def get_device_name(self, device_id):
        if self._devices_by_id is None:
            self._read_devices()
        return self._devices_by_id.get(device_id)

    def get_device_id(self, name):
        if self._devices_by_id is None:
            self._read_devices()
        return self._devices_by_name.get(name)

    @property
    def device_file(self):
        return os.path.join(self.data_dir, 'devices.txt')

    def _read_devices(self):
        self._devices_by_name = {}
        pat = '^(?!#)(.+)=\\s(.+)'
        try:
            with open(self.device_file, 'r') as f:
                for line in f:
                    if not line.startswith('#'):
                        try:
                            name, code = line.rsplit('=', 1)
                            self._devices_by_name[name.strip()] = code.strip()

                        except Exception as e:
                            pass
            self._devices_by_id = {
                code: name for name, code in self._devices_by_name.items()
            }

        except:
            self._devices_by_id = {}

    def _write_devices(self):
        header = '\n'.join([
            '# Names for devices based on their device ids, format:',
            '#',
            '#  name = code',
            '#',
            '# name is space trimmed, but can contain spaces',
            '# Blank lines are ignored. Codes and names should be unique.',
            '',
        ])

        with open(self.device_file, 'w+') as f:
            print(header, file=f)
            for code, name in self._devices_by_id.items():
                print('{0:20s} = {1}'.format(name, code), file=f)


class StreamWriter(object):
    def __init__(self, writer, progress_fun=None):
        self._w = writer
        self._bytes_written = 0
        self._start_time = timeit.default_timer()
        self._latest_time = self._start_time
        self._progress_fun = progress_fun

    @property
    def bytes_written(self):
        return self._bytes_written

    def write(self, b):
        count = self._w.write(b)
        self._latest_time = timeit.default_timer()
        self._bytes_written += count
        if self._progress_fun:
            self._progress_fun(count, self)
        return count

    def report(self):
        elapsed = self._latest_time - self._start_time
        rate = round(float(self._bytes_written) / 1000 / elapsed, 1)
        return 'received {} bytes in {} seconds, rate: {} KB/s'.format(
            self._bytes_written, round(elapsed, 2), rate)



import os
import locale
import datetime
import time
import urwid as uw
from urwid_timed_progress import TimedProgressBar


class App(object):

    palette = [
        ('normal',   'white', 'black', 'standout'),
        ('complete', 'white', 'dark magenta'),
        ('footer',   'white', 'dark gray'),
    ]

    # Using SI units: https://en.wikipedia.org/wiki/Kilobyte
    units = [
        ('B', 1),
        ('kB', 1000),
        ('MB', 1000000),
        ('GB', 1000000000),
    ]

    locale.setlocale(locale.LC_ALL, '')

    def __init__(self, reaper, rm_after_download=True, auto_run=False):
        self.rm_after_download = rm_after_download
        self.reaper = reaper
        self.file_progress = TimedProgressBar('normal',
                                              'complete',
                                              label='Current File',
                                              label_width=13,
                                              units=App.units,
                                              done=.001)

        self.overall_progress = TimedProgressBar('normal',
                                                 'complete',
                                                 label='Overall',
                                                 label_width=13,
                                                 units=App.units,
                                                 done=.001)

        info = self.reaper.info()
        self.device_id = info['samd_id'][2:]
        self.file_count = uw.Text('')
        self.file_list = self.reaper.ls()
        self.update_file_count()

        info_text = [
            'Device Id: {}'.format(self.device_id),
            'Port: {}'.format(self.reaper.conn.name),
            'SD Info: {} {} {} {} {}'.format(
                info['sd_mid'][2:],
                info['sd_oid'],
                info['sd_pnm'],
                info['sd_prv'],
                info['sd_mdt']),
            'SD Serial: {}'.format(info['sd_psn'][2:]),
            'SD Size: {:.2f} GB'.format(float(info['sd_size']) * 512 / 1e9),
        ]

        # Lookup device name, create a short one if not found
        device_name = self.reaper.get_device_name(self.device_id)
        if device_name is None:
            device_name = 'tag-{}'.format(self.device_id[-5:])
            self.reaper.set_device_name(device_name, self.device_id)

        # Invoked when editing device name is complete
        def device_name_handler(new_name):
            device_id = self.reaper.get_device_id(new_name)

            # name not changed
            if device_id == self.device_id:
                self.status('-ready-')
                return True

            # name not being used yet
            if device_id is None:
                self.reaper.set_device_name(new_name, self.device_id)
                self.status('-ready-')
                return True

            # name already being used
            else:
                msg = ('Device Name "{}" already used for device "{}"\n' +
                       'Please choose another name.').format(new_name,
                                                             device_id)
                self.status(msg)
                return False

        # device name edit field
        self.device_name = edit_field('Device Name: ',
                                      device_name,
                                      device_name_handler)

        info_left = uw.Pile(
            [self.device_name] +
            [uw.Text(t) for t in info_text] +
            [self.file_count])

        self.clock = Clock('')

        info_right = uw.Pile([uw.Padding(self.clock, align='right', width=13)])

        info_box = uw.LineBox(uw.Columns([info_left, info_right]))

        self._status = uw.Text('-ready-')
        self.debug_message = uw.Text('')

        commands = [('r', 'run download'),
                    ('n', 'set device name'),
                    ('c', 'set device clock'),
                    ('q', 'exit')]

        footer = uw.Text(' | '.join(['{} = {}'.format(key, text)
                                    for key, text in commands]))

        self.footer = uw.AttrWrap(footer, 'footer')

        main_view = uw.Frame(uw.ListBox([info_box,
                                         uw.Divider(),
                                         self.file_progress,
                                         uw.Divider(),
                                         uw.Divider(),
                                         self.overall_progress,
                                         uw.Divider(),
                                         uw.Divider(),
                                         self._status,
                                         uw.Divider(),
                                         self.debug_message,
                                         ]),
                             footer=self.footer)

        def keypress(key):
            if key in ('q', 'Q'):
                raise uw.ExitMainLoop()
            elif key in ('r', 'R'):
                self.download()
            elif key in ('c', 'C'):
                self.set_device_clock()
            elif key in ('n', 'N'):
                self.status('Enter new Device Name')
                self.device_name.edit_mode()

        self.loop = uw.MainLoop(main_view,
                                App.palette,
                                unhandled_input=keypress)
        if auto_run:
            def run(*args, **kwargs):
                self.set_device_clock()
                time.sleep(.5)
                self.download()
            self.loop.set_alarm_in(0.5, run)
        else:
            self.loop.set_alarm_in(0.5, self.set_device_clock)

        self.loop.set_alarm_in(1, self.clock.refresh)
        self.loop.run()

    def set_device_clock(self, *args, **kwargs):
        self.status('Setting device clock to current UTC time ...')
        self.reaper.set_time()
        dt = self.reaper.get_time()
        self.status('Device clock set to {}\n\n-ready-'.format(dt))

    def status(self, msg):
        self._status.set_text(msg)
        self.clock.refresh_once()
        self.loop.draw_screen()

    def update_file_count(self):
        if self.file_list is None:
            t = 'n/a'
        else:
            t = locale.format('%d', len(self.file_list), grouping=True)
        self.file_count.set_text('# of Files: {}'.format(t))

    def download(self):
        if self.file_list is None:
            self.get_file_list()
        total_size = sum([f['size'] for f in self.file_list])
        self.overall_progress.done = total_size if total_size != 0 else .001
        self.status('starting download ...')
        for f in list(self.file_list):
            if f['size'] > 0:
                self.file_progress.reset()
                self.file_progress.done = f['size']
                self.loop.draw_screen()

                def progress_fun(num_bytes, _):
                    self.file_progress.add_progress(num_bytes)
                    self.overall_progress.add_progress(num_bytes)
                    self.clock.refresh_once()
                    self.loop.draw_screen()

                data_dir = os.path.join(self.reaper.data_dir, self.device_id)

                sd_filename = f['name'].lstrip('/')
                local_filename = os.path.join(data_dir, sd_filename)
                dirname = os.path.dirname(local_filename)
                if not os.path.exists(dirname):
                    os.makedirs(dirname)

                while os.path.exists(local_filename):
                    local_filename = increment_filename(local_filename)

                self.status('downloading {} to {} ...'.format(sd_filename,
                                                              local_filename))
                self.reaper.cp(sd_filename,
                               local_filename,
                               f['size'],
                               progress_fun)

            # remove file
            if self.rm_after_download:
                self.reaper.rm(f['name'])
            self.file_list = self.file_list[1:]
            self.update_file_count()
        self.status('-ready-')

    def debug(self, m):
        self.debug_message.set_text(m)
        self.loop.draw_screen()


def edit_field(caption, edit_text, handler):
    edit = uw.Edit(caption, edit_text)
    ef = EditField(edit, handler)
    return uw.BoxAdapter(ef, height=1)


class Clock(uw.Text):
    def refresh_once(self):
        now = datetime.datetime.utcnow()
        self.set_text('{}\n{}'.format(
            now.strftime('%Y-%m-%d'),
            now.strftime('%H:%M:%S UTC')))

    def refresh(self, loop=None, data=None):
        self.refresh_once()
        loop.set_alarm_in(1, self.refresh)


class EditField(uw.Filler):
    def __init__(self, edit, handler):
        super(EditField, self).__init__(edit)
        self.edit_widget = edit
        self.handler = handler
        self.text_widget = uw.Text('')
        self.value = self.edit_widget.edit_text
        self.text_mode()

    def edit_mode(self):
        self.original_widget = self.edit_widget

    def text_mode(self):
        self.text_widget.set_text(self.edit_widget.caption + self.value)
        self.original_widget = self.text_widget

    def keypress(self, size, key):
        if key == 'esc':
            self.text_mode()
            self.handler(self.value)

        elif key == 'enter':
            if self.edit_widget.edit_text != self.value:
                if self.handler(self.edit_widget.edit_text):
                    self.value = self.edit_widget.edit_text
                    self.text_mode()
        else:
            return super(EditField, self).keypress(size, key)


def increment_filename(filename):
    parts = filename.rsplit('-', 1)
    try:
        counter = int(parts[1])
    except:
        counter = 0
    return '{}-{}'.format(parts[0], counter + 1)



#!/usr/bin/env python

"""
Usage: sdreaper [options] [<command>...]

Examples:
    sdreaper
    sdreaper -d c:\\tag-data

Options:
  -r --run                 begin download automatically
  -d DATA --data=DATA      data directory to use [default: data]

Advanced Options:
  -p PORT --port=PORT      Port to use. By default the port is auto detected.
  --monitor                monitor and print serial output
  --get-time               get device time, assumed to be UTC
  --set-time               set device time to current system time, in UTC
  --find-devices           list connected Arduino Reaper boards
  --no-rm                  do not remove files after download, by default
                           files are removed after they are downloaded
  -h --help                show help
  --version                show version
"""

import sys
from docopt import docopt
from sdreaper.reaper import Reaper
from sdreaper.app import App
import logging

logging.basicConfig(format='%(levelname)-5s %(message)s',
                    filename='sdreaper.log',
                    level=logging.INFO)


def main():
    args = docopt(__doc__, version='sdreaper 1.0.1')

    if args['--find-devices']:
        print('Finding devices...\n')
        for p in Reaper.find():
            print(p)
        return

    port = args.get('--port')
    if not port:
        try:
            port = Reaper.find()[0]
        except IndexError:
            print('Unable to auto-detect device. Please check connection.\n'
                  'If device is connected you may need to specify port'
                  ' with -p')
            return 1

    reaper = Reaper(port=port, data_dir=args['--data'])
    reaper.connect()

    if args['--monitor']:
        while True:
            reaper.read()
    elif args['--get-time']:
        reaper.echo = False
        print(reaper.get_time())
    elif args['--set-time']:
        reaper.echo = False
        print(reaper.set_time())
    elif args['<command>']:
        reaper.commands(args['<command>'])
    else:
        reaper.echo = False
        App(reaper, not args['--no-rm'], args['--run'])

    reaper.disconnect()

if __name__ == '__main__':
    sys.exit(main())



"""
|Downloads|

sdreaper
========

CLI that communicates with Arduino Reaper to pull files off an SD card and
delete them using a Serial connection.

.. |Downloads| image:: https://img.shields.io/pypi/dm/sdreaper.svg
   :target: https://pypi.python.org/pypi/sdreaper

"""
import os
import sys

from setuptools import setup, find_packages

setup(
    name='sdreaper',
    version='1.0.1',
    description='CLI and curses UI for talking to Arduino Reaper library',
    long_description=__doc__,
    url='https://github.com/humulabs/reaper/blob/master/README.md',
    author='Michael Keirnan',
    author_email='michael@keirnan.com',
    packages=['sdreaper'],
    package_dir={'': '.'},
    install_requires=[
        'pyserial',
        'docopt',
        'urwid_timed_progress',
    ],
    entry_points={
        'console_scripts': ['sdreaper=sdreaper.main:main'],
    },
    tests_require=[
        'mock',
        'pep8',
        'pytest',
    ],
    platforms='any',
    license='MIT',
    keywords="arduino iot sd serial",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Topic :: Home Automation',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ]
)
